﻿using System.Collections.Concurrent;
using System.Collections.Specialized;
using System.Net;
using System.Net.Http.Headers;

namespace SddpViewer;

using System.Collections.ObjectModel;
using System.Threading.Tasks;
using Avalonia.Threading;
using CommunityToolkit.Mvvm.ComponentModel;
using CommunityToolkit.Mvvm.Input;
using Rssdp;

public sealed partial class MainWindowViewModel : ObservableObject, IDisposable
{
  public MainWindowViewModel()
  {
    NetworkAdapters = NetworkAdapter.GetAvailableNetworkAdapter().ToArray();
    SelectedNetworkAdapter = NetworkAdapters.FirstOrDefault(x => x.IpAddress.ToString() == DeviceIpAddress);
    SddpDevices = [];
    SddpDevices.CollectionChanged += SddpDevices_OnCollectionChanged;
  }

  private static async void SddpDevices_OnCollectionChanged(object? sender, NotifyCollectionChangedEventArgs e)
  {
    try
    {
      switch (e.Action)
      {
        case NotifyCollectionChangedAction.Add:
          await AddAction(e.NewItems);
          break;
        case NotifyCollectionChangedAction.Remove:
          RemoveAction(e.OldItems);
          break;
        case NotifyCollectionChangedAction.Replace:
          RemoveAction(e.OldItems);
          await AddAction(e.NewItems);
          break;
        case NotifyCollectionChangedAction.Reset:
          RemoveAction(sender as ObservableCollection<DiscoveredDeviceViewModel>);
          break;
        case NotifyCollectionChangedAction.Move:
          break;
        default:
          throw new ArgumentOutOfRangeException();
      }
    }
    catch (Exception ex)
    {
      Console.WriteLine(ex.Message);
    }
  }

  private static void RemoveAction(IList? list)
  {
    foreach (object eOldItem in list ?? Array.Empty<object>())
    {
      if (eOldItem is DiscoveredDeviceViewModel discoveredDeviceViewModel)
      {
        discoveredDeviceViewModel.Dispose();
      }
    }
  }

  private static async Task AddAction(IList? list)
  {
    foreach (object eNewItem in list ?? Array.Empty<object>())
    {
      if (eNewItem is DiscoveredDeviceViewModel discoveredDeviceViewModel)
      {
        await discoveredDeviceViewModel.GetFurtherInformationAsync();
      }
    }
  }

  private void LocatorOnDeviceUnavailable(object? sender, DeviceUnavailableEventArgs e)
  {
    var existingDevice = SddpDevices.FirstOrDefault(x => string.Equals(x.Usn, e.DiscoveredDevice.Usn, StringComparison.Ordinal));
    if (existingDevice is null)
      return;
    Dispatcher.UIThread.Invoke(() => SddpDevices.Remove(existingDevice));
  }

  private void LocatorOnDeviceAvailable(object? sender, DeviceAvailableEventArgs e)
  {
    if (!e.IsNewlyDiscovered)
    {
      return;
    }

    var discoveredDeviceViewModel = new DiscoveredDeviceViewModel(e.DiscoveredDevice);
    if (!SddpDevices.Contains(discoveredDeviceViewModel))
      Dispatcher.UIThread.Invoke(() => SddpDevices.Add(discoveredDeviceViewModel));
  }

  [ObservableProperty]
  public partial IReadOnlyList<NetworkAdapter> NetworkAdapters { get; set; }

  [ObservableProperty]
  public partial NetworkAdapter? SelectedNetworkAdapter { get; set; }

  [ObservableProperty]
  public partial string DeviceIpAddress { get; set; } = "192.168.42.192";

  [ObservableProperty]
  public partial string NotificationFilter { get; set; } = "upnp:rootdevice";

  private SsdpDeviceLocator? _locator;

  public ObservableCollection<DiscoveredDeviceViewModel> SddpDevices { get; }

  [RelayCommand]
  private async Task SearchDevicesNowAsync()
  {
    SddpDevices.Clear();
    if (_locator is not null)
    {
      _locator.StopListeningForNotifications();
      _locator.DeviceAvailable -= LocatorOnDeviceAvailable;
      _locator.DeviceUnavailable -= LocatorOnDeviceUnavailable;
      _locator.Dispose();
    }

    await StartListening(async locator =>
    {
      foreach (var discoveredSsdpDevice in await locator.SearchAsync())
      {
        var discoveredDeviceViewModel = new DiscoveredDeviceViewModel(discoveredSsdpDevice);
        Dispatcher.UIThread.Invoke(() => SddpDevices.Add(discoveredDeviceViewModel));
      }
    });
  }

  [RelayCommand]
  private async Task ResearchAsync()
  {
    if (_locator is not null)
    {
      await _locator.SearchAsync();
    }
  }

  [RelayCommand(IncludeCancelCommand = true)]
  private async Task IpScan(CancellationToken cancellationToken)
  {
    if (SelectedNetworkAdapter is not null)
    {
      var ip = SelectedNetworkAdapter.IpAddress.GetAddressBytes();
      ip[^1] = 0;
      var interNetwork = new IPAddress(ip);
      var cidr = $"{interNetwork}/{SelectedNetworkAdapter.PrefixLength}";
      var range = CidrHelper.GetIpRangeFromCidr(cidr);
      var ipRange = IpRangeHelper.GetIPRange(range.Start, range.End).Where(address => IpAddressFilter(address, range));
      var collection = new ConcurrentBag<DiscoveredDeviceViewModel>();
      using var httpClient = new HttpClient();
      httpClient.Timeout = TimeSpan.FromSeconds(1);

      await Parallel.ForEachAsync(
        ipRange,
        new ParallelOptions() { MaxDegreeOfParallelism = 16, CancellationToken = cancellationToken },
        async (address, token) =>
        {
          try
          {
            token.ThrowIfCancellationRequested();
            DiscoveredSsdpDevice device = new()
            {
              CacheLifetime = TimeSpan.Zero,
              DescriptionLocation = new Uri($"http://{address.ToString()}:8080/discovery/ssdp"),
            };
            var response = await httpClient.GetAsync(device.DescriptionLocation, token);
            response.EnsureSuccessStatusCode();
            device.ResponseHeaders = response.Headers;
            var viewModel = new DiscoveredDeviceViewModel(device);
            collection.Add(viewModel);
            Console.WriteLine($"{address} ok");
          }
          catch (Exception)
          {
            Console.WriteLine($"{address} not ok");
          }
        }
      );

      Dispatcher.UIThread.Invoke(() =>
      {
        foreach (DiscoveredDeviceViewModel viewModel in collection)
        {
          SddpDevices.Add(viewModel);
        }
      });
    }
  }

  private bool IpAddressFilter(IPAddress arg, (IPAddress Start, IPAddress End) range)
  {
    if (Equals(arg, range.Start))
      return false;
    if (Equals(arg, range.End))
      return false;
    if (this.SddpDevices.Any(x => Equals(x.IpAddress, arg)))
      return false;
    return true;
  }

  [RelayCommand]
  private async Task StartListening(Func<SsdpDeviceLocator, Task>? action = null)
  {
    _locator = new SsdpDeviceLocator(SelectedNetworkAdapter?.IpAddress.ToString());
    if (!string.IsNullOrWhiteSpace(NotificationFilter))
    {
      _locator.NotificationFilter = NotificationFilter;
    }

    if (action is not null)
    {
      await action.Invoke(_locator);
    }

    _locator.DeviceAvailable += LocatorOnDeviceAvailable;
    _locator.DeviceUnavailable += LocatorOnDeviceUnavailable;
    _locator.StartListeningForNotifications();
  }

  public void Dispose()
  {
    _locator?.Dispose();
  }
}
